#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define __USE_GNU
//#define __USE_BSD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fnmatch.h>
#include <errno.h>

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "utils.h"

static int opReplace, opRecursive, isResultsVisible=0, clipboardPage=1;
static gboolean fsOpen_ack, fsBrowse_ack;

extern void *memmem (__const void *__haystack, size_t __haystacklen,
                     __const void *__needle, size_t __needlelen);

typedef struct {
	gchar *filename;
	guint matches;
	gshort status; /* OK=0, failed=-1 */
} fileList;

void forceGtkIdle ()
{
	while ( gtk_events_pending () )
	gtk_main_iteration ();
//	gdk_flush ();
}

void pushStatus ( int cid, char *str )
{
	gtk_statusbar_push ( GTK_STATUSBAR (sBar), cid, str);
}

void pushStatusStrip ( int cid, char *s)
{
	int n, i;
	gint sbWidth, textWidth;
	gchar *str = s;
	
	
	/* prevent resizing when statusBar's label is too long */
	textWidth = gdk_string_width ( GTK_STATUSBAR (sBar) -> label -> style -> font,
			str );
	n = gdk_string_width ( GTK_STATUSBAR (sBar) -> label -> style -> font, "M" );
	sbWidth = GTK_STATUSBAR (sBar) -> frame -> allocation.width - 2*n;
	i = 0;
	if ( textWidth > sbWidth ){
		/* if so, let's wipe out first chars */
		while ( gdk_string_width ( GTK_STATUSBAR (sBar) -> label -> style -> font,
					str ) > sbWidth ){
			str += 2;
			i += 2;
		}
	}
	
	pushStatus ( cid, str );
}

void popStatus ( int cid )
{
	gtk_statusbar_pop ( GTK_STATUSBAR (sBar), cid );
}

int openFile(filePage *fp, gchar *filename)
{
	gchar *buf;
	FILE *f;
	struct stat st;
	off_t size, maxsize, read, i, limit, input;
	guint cid;
	gchar msg[300];

	if ( stat( filename, &st)){
		g_error ("could not stat %s\n", filename);
		return -1;
	}
	size = st.st_size;
	if (size < 65536)
		limit = st.st_blksize;
	else if (size < 1048576)
		limit = 65536;
	else limit = 131072;
	if ( size <= limit )
		maxsize = size;
	else maxsize = limit;
	buf = g_malloc ( maxsize );
	gtk_text_freeze ( GTK_TEXT (fp -> text ));
	f = fopen ( filename, "r" );
	fflush (f);
	gtk_progress_configure ( GTK_PROGRESS (pBar), 0, 0, size);
	gtk_progress_bar_set_bar_style ( GTK_PROGRESS_BAR (pBar),
			GTK_PROGRESS_DISCRETE);
	read = 0;
	sprintf ( msg, _(" Loading %ld bytes from %s ..."), size, filename );
	cid = gtk_statusbar_get_context_id ( GTK_STATUSBAR (sBar), "load");
	pushStatus ( cid, msg );
	while ( (read < size) && !feof(f) ){
		i = fread ( buf, maxsize, 1, f);
		if (i) input = maxsize;
		else{
			/* test for io error */
			input = size - read;
		}
		read += input;
		gtk_progress_set_value ( GTK_PROGRESS (pBar), read);
		forceGtkIdle ();
		/* we can use font here (1st NULL) */
		gtk_text_insert ( GTK_TEXT (fp -> text),
				NULL, NULL, NULL, buf, input);
	}
	fclose (f);
	
	gtk_progress_set_value ( GTK_PROGRESS (pBar), read);
	forceGtkIdle ();
	gtk_text_thaw ( GTK_TEXT ( fp -> text ));
	gtk_progress_set_value ( GTK_PROGRESS (pBar), 0);
	popStatus ( cid );
	return 0;
}

filePage *newFilePage( gchar *filename)
{
	filePage *fp;
	GtkWidget *frame, *scroller, *txt;
	
	fp = g_malloc ( sizeof (filePage) );
	frame = gtk_frame_new (NULL);
	fp -> frame = frame;
	
	scroller = gtk_scrolled_window_new( NULL, NULL);
	fp -> scroller = scroller;
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW (scroller),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add ( GTK_CONTAINER (frame), scroller);
	
	txt=gtk_text_new(NULL, NULL);
	fp->text=txt;
	gtk_text_set_editable(GTK_TEXT(txt), FALSE);
	gtk_container_add ( GTK_CONTAINER (scroller),	txt);
	
	fp -> filename = g_strdup( filename );

	openFile ( fp, fp -> filename );
	
	return fp;
}

void fpDestroy ( GtkWidget *w, gpointer data )
{
	filePage *fp = gtk_object_get_data ( GTK_OBJECT (w), "fp");
//	g_print ("destroy data: %s\n", fp -> filename);
	g_free (fp -> filename);
	g_free (fp);
}

void cb_closePage ( GtkWidget *w, gpointer data )
{
	gint page = gtk_notebook_page_num (GTK_NOTEBOOK (notebook1),
			GTK_WIDGET (data));
	fpDestroy ( data, NULL);
	gtk_notebook_remove_page ( GTK_NOTEBOOK (notebook1), page );
	clipboardPage--;
}

void insertFilePage( filePage *fp )
{
	gchar *name;
	GtkWidget *label, *hbox, *but, *pix;

	/* get the filename (w/o path, just like basename) */
	name = strrchr( fp -> filename, '/' );
	name ++;

	hbox = gtk_hbox_new (FALSE, 3);
	but = gtk_button_new ();
	pix = create_pixmap ( notebook1, "close.xpm");
	gtk_container_add ( GTK_CONTAINER (but), pix);
	label = gtk_label_new ( name );
	gtk_box_pack_start( GTK_BOX (hbox), but, FALSE, FALSE, 0);
	gtk_box_pack_start( GTK_BOX (hbox), label, FALSE, FALSE, 0);
	gtk_widget_show ( hbox );
	gtk_widget_show ( label );
	gtk_widget_show ( pix );
	gtk_button_set_relief ( GTK_BUTTON (but), GTK_RELIEF_NONE );
	gtk_widget_show ( but );
	gtk_signal_connect ( GTK_OBJECT (but), "clicked",
			GTK_SIGNAL_FUNC (cb_closePage), fp -> frame);
	
	gtk_notebook_insert_page_menu ( GTK_NOTEBOOK (notebook1),
			fp -> frame, hbox, gtk_label_new ( fp -> filename ),
			clipboardPage++);
	gtk_widget_show ( fp -> text );
	gtk_widget_show ( fp -> scroller );
	gtk_frame_set_label ( GTK_FRAME (fp -> frame), fp -> filename );
	gtk_widget_show ( fp -> frame );

	gtk_object_set_data ( GTK_OBJECT (fp -> frame), "fp", fp);
}

void cb_fsOpen_ok(GtkWidget *w, gpointer data)
{
	gchar *filename;
	filePage *fp;
	
	filename = g_strdup ( gtk_file_selection_get_filename (GTK_FILE_SELECTION (data)));
	gtk_widget_destroy ( GTK_WIDGET (data));
	fsOpen_ack=FALSE;
	fp = newFilePage ( filename );
	g_free ( filename );
	insertFilePage ( fp );
	gtk_notebook_set_page ( GTK_NOTEBOOK (notebook1), clipboardPage-1 );
}

void cb_fsOpen_cancel(GtkWidget *w, gpointer data)
{
	gtk_widget_destroy ( GTK_WIDGET(data));
	fsOpen_ack=FALSE;
}

void cb_Open( GtkWidget *w, gpointer data)
{
	GtkWidget *fs;
	
	if ( !fsOpen_ack ){
		fs = gtk_file_selection_new ( _("Open File") );
		gtk_signal_connect ( GTK_OBJECT(fs), "destroy", 
			GTK_SIGNAL_FUNC (cb_fsOpen_cancel), fs);
		gtk_signal_connect ( GTK_OBJECT ( GTK_FILE_SELECTION (fs) -> ok_button),
			"clicked", GTK_SIGNAL_FUNC (cb_fsOpen_ok), fs);
		gtk_signal_connect ( GTK_OBJECT( GTK_FILE_SELECTION (fs) -> cancel_button),
				"clicked", GTK_SIGNAL_FUNC ( cb_fsOpen_cancel),
				fs);
		gtk_window_set_modal ( &(GTK_FILE_SELECTION (fs) -> window),
				TRUE);
		gtk_widget_show ( fs );
		fsOpen_ack = TRUE;
	}
}

void cb_ExitForSure( GtkWidget *w, gpointer data)
{
	gtk_exit ( 0 );
}

void msgBox2but ( gchar *title, gchar *msg, GtkWidget *b1, GtkWidget *b2 );
void delWindow ( GtkWidget *w, GtkWidget *window );

void cancelExit ( GtkWidget *w, gpointer data )
{
	char path[200];
	
	sprintf( path, "esdplay %s/sounds/turma-lb-fuck.wav", PACKAGE_DATA_DIR);
	system ( path );
	delWindow ( w, NULL );
}

void cb_Exit ( GtkWidget *w, gpointer data )
{
	GtkWidget *bExitOk, *bExitCancel;

	bExitOk = gtk_button_new_with_label ( _("Please") );
	bExitCancel = gtk_button_new_with_label ( _("Fsck you!") );
	gtk_signal_connect ( GTK_OBJECT ( bExitOk ), "clicked",
			GTK_SIGNAL_FUNC ( cb_ExitForSure ), NULL );
	gtk_signal_connect ( GTK_OBJECT (bExitCancel), "clicked",
		GTK_SIGNAL_FUNC (cancelExit), NULL );
	msgBox2but ( _("The way out"),
			_("I dunno man, you have a funny nose.\n You have to say please...\n"
			"But, as usual, you have another option."),
			bExitOk, bExitCancel );
}

void cb_scrollEntry ( GtkWidget *w, gpointer data )
{
	gtk_entry_set_position ( GTK_ENTRY (w),
			strlen ( gtk_entry_get_text ( GTK_ENTRY (w) ) ));
}

void cb_fsBrowse_ok(GtkWidget *w, gpointer data)
{
	gchar *filename;
	GList *glist = NULL;
	GtkWidget *li;
	
	filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (data));
	li = gtk_list_item_new_with_label (filename);
	gtk_widget_show (li);
	glist = g_list_append ( glist,  li);
	gtk_list_prepend_items ( GTK_LIST ( GTK_COMBO (comboBrowse) -> list ),
			glist);
	gtk_list_select_item ( GTK_LIST ( GTK_COMBO (comboBrowse) -> list ), 0);
	gtk_widget_destroy( GTK_WIDGET (data));
	fsBrowse_ack = FALSE;
}

void cb_fsBrowse_cancel(GtkWidget *w, gpointer data)
{
	gtk_widget_destroy( GTK_WIDGET(data));
	fsBrowse_ack = FALSE;
}

void cb_dirToggle ( GtkWidget *w, gpointer data )
{
	gtk_widget_set_sensitive ( GTK_WIDGET (data),
			GTK_TOGGLE_BUTTON (w) -> active);
}

void cb_Browse( GtkWidget *w, gpointer data)
{
	GtkWidget *fs, *dirToggle;
	
	if(!fsBrowse_ack){
		fs = gtk_file_selection_new ( _("Browse") );
		gtk_signal_connect( GTK_OBJECT(fs), "destroy", 
			GTK_SIGNAL_FUNC(cb_fsBrowse_cancel), fs);
		gtk_signal_connect( GTK_OBJECT( GTK_FILE_SELECTION (fs) -> ok_button),
			"clicked", GTK_SIGNAL_FUNC(cb_fsBrowse_ok), fs);
		gtk_signal_connect( GTK_OBJECT( GTK_FILE_SELECTION (fs) -> cancel_button),
				"clicked", GTK_SIGNAL_FUNC( cb_fsBrowse_cancel),
				fs);
		gtk_widget_set_sensitive ( GTK_FILE_SELECTION (fs) -> file_list, FALSE );
		dirToggle = gtk_check_button_new_with_label ( _("Select files") );
		gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON (dirToggle), FALSE);
		gtk_signal_connect ( GTK_OBJECT (dirToggle), "toggled",
				GTK_SIGNAL_FUNC (cb_dirToggle),
				GTK_FILE_SELECTION (fs) -> file_list);
		gtk_box_pack_start ( GTK_BOX ( GTK_FILE_SELECTION (fs) -> button_area ),
				dirToggle, TRUE, TRUE, 10 );
		gtk_box_set_child_packing ( GTK_BOX ( GTK_FILE_SELECTION (fs) -> main_vbox ),
				GTK_FILE_SELECTION (fs) -> button_area,
				TRUE, TRUE, 10, GTK_PACK_START );
		gtk_window_set_modal ( &(GTK_FILE_SELECTION (fs) -> window),
				TRUE);
		gtk_widget_show (dirToggle);
		gtk_widget_show (fs);
		fsBrowse_ack = TRUE;
	}
}

void cb_Recursive( GtkWidget *w, gpointer data)
{
	opRecursive = GTK_TOGGLE_BUTTON(data) -> active;
}

void cb_Replace( GtkWidget *w, gpointer data)
{
	opReplace = GTK_TOGGLE_BUTTON(data) -> active;
	if( tReplace )
		gtk_widget_set_sensitive ( tReplace, opReplace );
}

void delWindow ( GtkWidget *w, GtkWidget *window )
{
	GtkWidget *dlg = window;
	
	if ( !dlg ){
		dlg = w -> parent -> parent -> parent -> parent;
	}
	gtk_widget_destroy ( dlg );
}

void msgBox ( gchar *title, gchar *msg, gchar *buttonLabel )
{
	GtkWidget *button, *label, *dlg, *pix, *hbox, *bbox;

	dlg = gtk_dialog_new ();
	gtk_container_set_border_width ( GTK_CONTAINER ( GTK_DIALOG (dlg) -> vbox), 5);
	gtk_container_set_border_width ( GTK_CONTAINER ( GTK_DIALOG (dlg) -> action_area), 5);
	gtk_window_set_title ( &(GTK_DIALOG (dlg) -> window ),
			title);
	gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
			    GTK_SIGNAL_FUNC (gtk_widget_destroy), NULL);

	gtk_signal_connect (GTK_OBJECT (dlg), "delete_event",
			    GTK_SIGNAL_FUNC (gtk_widget_destroy), NULL);

	hbox = gtk_hbox_new ( FALSE, 10);
	gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (dlg) -> vbox ),
			hbox, TRUE, TRUE, 10);
	gtk_widget_show (hbox);

	pix = create_pixmap ( dlg, "msg.xpm" );
	gtk_box_pack_start ( GTK_BOX ( hbox ), pix, FALSE, FALSE, 0);
	gtk_widget_show (pix);
	
	label = gtk_label_new (msg);
	gtk_box_pack_start ( GTK_BOX ( hbox ), label, TRUE, TRUE, 0);
	gtk_widget_show (label);
	
	bbox = gtk_hbutton_box_new ();
	gtk_button_box_set_layout( GTK_BUTTON_BOX ( bbox ), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX ( bbox ), 5);
	gtk_button_box_set_child_ipadding_default ( 5, 5 );
	gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (dlg) -> action_area ),
			bbox, TRUE, TRUE, 5);
	gtk_widget_show ( bbox );
	
	button = gtk_button_new_with_label ( buttonLabel );
	gtk_signal_connect ( GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (delWindow), dlg);
	gtk_container_add ( GTK_CONTAINER ( bbox ), button );
	gtk_widget_show ( button );
	GTK_WIDGET_SET_FLAGS ( button, GTK_CAN_DEFAULT );
	gtk_widget_grab_default ( button );
	gtk_widget_grab_focus ( button );

	gtk_widget_show ( dlg );
	gtk_window_set_modal ( &(GTK_DIALOG (dlg) -> window ), TRUE);
}

void msgBox2but ( gchar *title, gchar *msg, GtkWidget *b1, GtkWidget *b2 )
{
	GtkWidget *label, *dlg, *pix, *hbox, *bbox;

	dlg = gtk_dialog_new ();
	gtk_container_set_border_width ( GTK_CONTAINER ( GTK_DIALOG (dlg) -> vbox), 5);
	gtk_container_set_border_width ( GTK_CONTAINER ( GTK_DIALOG (dlg) -> action_area), 5);
	gtk_window_set_title ( &(GTK_DIALOG (dlg) -> window ),
			title);
	gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
			    GTK_SIGNAL_FUNC (gtk_widget_destroy), NULL);

	gtk_signal_connect (GTK_OBJECT (dlg), "delete_event",
			    GTK_SIGNAL_FUNC (gtk_widget_destroy), NULL);

	hbox = gtk_hbox_new ( FALSE, 10);
	gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (dlg) -> vbox ),
			hbox, TRUE, TRUE, 10);
	gtk_widget_show (hbox);

	pix = create_pixmap ( dlg, "msg.xpm" );
	gtk_box_pack_start ( GTK_BOX ( hbox ), pix, FALSE, FALSE, 0);
	gtk_widget_show (pix);
	
	label = gtk_label_new (msg);
	gtk_box_pack_start ( GTK_BOX ( hbox ), label, TRUE, TRUE, 0);
	gtk_widget_show (label);
	
	bbox = gtk_hbutton_box_new ();
	gtk_button_box_set_layout( GTK_BUTTON_BOX ( bbox ), GTK_BUTTONBOX_END);
        gtk_button_box_set_spacing(GTK_BUTTON_BOX ( bbox ), 5);
	gtk_button_box_set_child_ipadding_default ( 5, 5 );
	gtk_box_pack_start ( GTK_BOX ( GTK_DIALOG (dlg) -> action_area ),
			bbox, TRUE, TRUE, 5);
	gtk_widget_show ( bbox );
	
	gtk_container_add ( GTK_CONTAINER ( bbox ), b1);
	GTK_WIDGET_SET_FLAGS ( b1, GTK_CAN_DEFAULT );
	gtk_widget_show (b1);
	gtk_container_add ( GTK_CONTAINER ( bbox ), b2);
	GTK_WIDGET_SET_FLAGS ( b2, GTK_CAN_DEFAULT );
	gtk_widget_show (b2);
	gtk_widget_grab_default ( b1 );
	gtk_widget_grab_focus ( b1 );

	gtk_widget_show (dlg);
	gtk_window_set_modal ( &(GTK_DIALOG (dlg) -> window ), TRUE);
}

static gchar *selectPattern;

int selectFiles ( struct dirent *dir )
{
	struct stat fstatus;
	
	if ( !strcmp ( dir -> d_name, "." ) || !strcmp ( dir -> d_name, ".." ) )
		return 0;
	
	if ( lstat ( dir -> d_name, &fstatus ) ){
		return 0;
	}

	if ( S_ISLNK ( fstatus.st_mode ) )
		/* ignore [sym]links for now (avoid recursivity,
		 * perhaps even infinite loops)
		 */
		return 0;

	if ( opRecursive && S_ISDIR ( fstatus.st_mode )){
		/* define some magic here to alert directories
		 * presence => one pass search.
		 * d_type appears to be defined only on BSD's
		 * so let's overide it with something usefull.
		 */
		dir -> d_type = 0xaa;
		return 1;
	}
	
	if ( S_ISREG ( fstatus.st_mode ) &&
			 !fnmatch ( selectPattern, dir -> d_name, 0 ) )
		return 1;

	/* the balast */
	return 0;
}

gint sortFiles ( gconstpointer a, gconstpointer b )
{
	fileList *e1, *e2;
	gchar *name1, *name2, *dir1, *dir2;
	gint res;

	e1 = (fileList*) a;
	e2 = (fileList*) b;
	name1 = g_basename ( e1 -> filename );
	name2 = g_basename ( e2 -> filename );
	dir1 = g_dirname ( e1 -> filename );
	dir2 = g_dirname ( e2 -> filename );

	res = strcmp ( dir1, dir2 );
	if ( !res )
		res = strcmp ( name1, name2 );
	
	g_free ( dir1 ); g_free ( dir2 );
	
	return res;
}

void insertFiles ( gpointer list, gint i );
typedef int (*scandir_selFunc)(const struct dirent *);
static int sbContext;
gboolean notCanceled = TRUE;

GSList *getFileList ( gchar *location, gchar *pattern )
{
	GSList *list = NULL;
	gchar *curdir;
	struct dirent **entry;
	fileList *fl;
	gfloat sbPerc;
	int n, i;

	if ( !notCanceled )
		return NULL;
	/* put a msg on statusbar about crt location */
	curdir = g_new ( gchar, strlen (location) + 22 );
	sprintf ( curdir, _("Making file list in %s"), location );
	pushStatusStrip ( sbContext, curdir );

	sbPerc = gtk_progress_get_current_percentage ( GTK_PROGRESS (pBar) );
	sbPerc += 0.10;
	if ( sbPerc > 1 )
		sbPerc = 0;
	gtk_progress_bar_update ( GTK_PROGRESS_BAR (pBar), sbPerc );
	forceGtkIdle ();
	g_free ( curdir );
	
	curdir = g_get_current_dir ();
	chdir ( location );

	selectPattern = pattern;
	n = scandir ( location, &entry, (scandir_selFunc) selectFiles, 0);

	chdir ( curdir );
	g_free ( curdir );
	
	if ( n >= 0 ){
		for ( i = 0; i < n; i++ ){
			if ( entry[i] -> d_type == 0xaa ){
				/* dir hit! let's look inside */
				curdir = g_strconcat ( location, entry[i] -> d_name, "/", NULL );
				list = g_slist_concat ( list, getFileList ( curdir, pattern ));
				g_free ( curdir );
				g_free ( entry[i] );
			}
			else {
				fl = g_new ( fileList, 1 );
				fl -> filename = g_strconcat ( location, entry[i] -> d_name, NULL );
				list = g_slist_append ( list, fl );
				g_free ( entry[i] );
			}
		}
		g_free ( entry );
	}

	popStatus ( sbContext );
	
	return list;
}

int doSearchReplace ( GSList *l, gchar *s, gchar *r, gint row )
{
	fileList *fl = (fileList*)(l -> data);
	FILE *f;
	char *buf, *ptr, *fbuf;
	struct stat filestat;
	unsigned long startPtr, crtPtr,
		slen=0, rlen=0,
		buflen, newlen;
	gchar str1[20], str2[20];
	int hits = 0;

	pushStatusStrip ( sbContext, fl -> filename );
	gtk_progress_set_value ( GTK_PROGRESS (pBar),
			1.0 + gtk_progress_get_value (GTK_PROGRESS (pBar)) );
	stat ( fl -> filename, &filestat );
	if ( (f = fopen (fl -> filename, "r")) ){
		buflen = filestat.st_size;
		newlen = buflen; /* newlen is the new filesize after replace */
		fbuf = g_malloc ( buflen );
		buf = fbuf;
		fread ( buf, buflen, 1, f );
		slen = strlen ( s );
		if ( r && opReplace )
			rlen = strlen ( r );
		do{
			ptr = (char*) memmem ( buf, buflen, s, slen );
			if ( ptr ){
				sprintf ( str1, "%p", fbuf );
				sprintf ( str2, "%p", ptr );
				startPtr = strtoul ( str1, NULL, 16 );
				crtPtr = strtoul ( str2, NULL, 16 );
				if ( rlen ){
					if ( slen < rlen ){
						newlen += ( rlen - slen );
						fbuf = g_realloc ( fbuf, newlen );
						buflen = newlen - ( crtPtr - startPtr + rlen );
						g_memmove ( ptr+rlen, ptr+slen, buflen );
						memcpy ( ptr, r, rlen );
						buf = ptr + rlen;
					}
					else if ( slen > rlen ){
						memcpy ( ptr, r, rlen );
						buflen = newlen - ( crtPtr - startPtr + slen);
						g_memmove ( ptr+rlen, ptr+slen, buflen );
						buf = ptr + rlen;
						newlen -= ( slen - rlen );
					}
					else {
						memcpy ( ptr, r, rlen );
						buf = ptr + slen;
						buflen = newlen - ( crtPtr - startPtr + slen );
					}
				}
				else{ /* search forward */
					buf = ptr + slen;
					buflen = newlen - ( crtPtr - startPtr + slen );
				}
				hits++;
			}
		} while ( ptr && (buflen >= slen) && notCanceled );
		/* let's write the new file */
		if ( hits && opReplace && r ){
			f = freopen ( fl -> filename, "w", f);
			fwrite ( fbuf, newlen, 1, f );
		}
		fclose (f);
		g_free ( fbuf );

		if ( hits ){
			fl -> matches = hits;
			/* failure presence should be reported here */
			fl -> status = 0;
			insertFiles ( fl, row );
		}
		
		forceGtkIdle ();
	}
	else{
		g_warning ("err fopen %s\n", fl -> filename );
	}
	popStatus ( sbContext );

	return hits;
}

void insertFiles ( gpointer list, gint i )
{
	fileList *l = (fileList*) list;
	GtkCList *cl = (GtkCList*) clist;
	gchar *name, *dir, *status, *row[5];
	gchar no[6];

	name = g_basename ( l -> filename );
	dir = g_dirname ( l -> filename );
	sprintf ( no, "%d", i);
	status = g_strdup_printf ( _("%d matches"), l -> matches );
	row[0] = no;
	row[1] = name;
	row[2] = dir;
	row[3] = opReplace ? _("Replace") : _("Search");
	row[4] = status;

	gtk_clist_append ( cl, row);
	g_free ( dir );
	g_free ( status );
//	gtk_clist_set_row_data ( cl, i, list );
}

void showResultsList ()
{
	GtkPacker *p = GTK_PACKER (packer2);
		
	if ( isResultsVisible )
		return;
	gtk_packer_add_defaults (p, frame1, GTK_SIDE_TOP,
			GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_PACK_EXPAND | GTK_FILL_X | GTK_FILL_Y));
	gtk_arrow_set (GTK_ARROW (arrow1), GTK_ARROW_DOWN, GTK_SHADOW_IN);
	gtk_packer_set_child_packing (p, scrolledwindow2, GTK_SIDE_TOP,
		GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_FILL_X | GTK_FILL_Y),
		p -> default_border_width,
		p -> default_pad_x, p -> default_pad_y,
		p -> default_i_pad_x, p -> default_i_pad_y );
	if ( opReplace )
		gtk_packer_set_child_packing (p, scrolledwindow3, GTK_SIDE_TOP,
			GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_FILL_X | GTK_FILL_Y),
			p -> default_border_width,
			p -> default_pad_x, p -> default_pad_y,
			p -> default_i_pad_x, p -> default_i_pad_y );
	else
		gtk_packer_set_child_packing (p, scrolledwindow3, GTK_SIDE_TOP,
			GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_FILL_X),
			p -> default_border_width,
			p -> default_pad_x, p -> default_pad_y,
			p -> default_i_pad_x, p -> default_i_pad_y );
	gtk_packer_reorder_child ( p, frame1, 4);
	isResultsVisible = TRUE;
}

void hideResultsList ()
{
	GtkPacker *p = GTK_PACKER (packer2);

	if ( !isResultsVisible )
		return;
	gtk_container_remove( GTK_CONTAINER (packer2), frame1);
	gtk_arrow_set (GTK_ARROW (arrow1), GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	gtk_packer_set_child_packing (p, scrolledwindow2, GTK_SIDE_TOP,
		GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_EXPAND | GTK_FILL_X | GTK_FILL_Y),
		p -> default_border_width,
		p -> default_pad_x, p -> default_pad_y,
		p -> default_i_pad_x, p -> default_i_pad_y );
	gtk_packer_set_child_packing (p, scrolledwindow3, GTK_SIDE_TOP,
		GTK_ANCHOR_NORTH_WEST, (GtkPackerOptions) (GTK_EXPAND | GTK_FILL_X | GTK_FILL_Y),
		p -> default_border_width,
		p -> default_pad_x, p -> default_pad_y,
		p -> default_i_pad_x, p -> default_i_pad_y );
	isResultsVisible = FALSE;
}

void switchResultsList ()
{
	if (!isResultsVisible)
		showResultsList ();
	else
		hideResultsList ();
}

void clearResultsList ()
{
	gtk_clist_clear ( GTK_CLIST (clist) );
}

void addPatternInComboList ()
{
	gchar *pattern;
	GList *glist = NULL;
	GtkWidget *li;

	/* FIXME: do not add existing entries */
	pattern = gtk_entry_get_text ( GTK_ENTRY (comboPattern_entry) );
	li = gtk_list_item_new_with_label ( pattern );
	gtk_widget_show ( li );
	glist = g_list_append ( glist,  li );
	gtk_list_prepend_items ( GTK_LIST ( GTK_COMBO (comboPattern) -> list ),
			glist);
	gtk_list_select_item ( GTK_LIST ( GTK_COMBO (comboPattern) -> list ), 0);
}

void cb_Go( GtkWidget *w, gpointer data)
{
	gchar *searchText = 0, *replaceText = 0,
		*location, *pattern = NULL, *tmp = NULL;
	gboolean have2freeLocation = FALSE;
	GSList *flist, *wlist;
	struct stat filestat;
	gint i;

	notCanceled = TRUE;
	/* check for search text */
	if ( gtk_text_get_length ( GTK_TEXT (tSearch) ) ){
		searchText  = gtk_editable_get_chars ( GTK_EDITABLE (tSearch),  0, -1);
	}
	else {
		msgBox ( _("I'm confused..."),
				_("'scuse my ignorance, but can you tell me again\n\"What the hell are you lookin' for?\""),
				_("Damn it!") );
		gtk_widget_grab_focus ( tSearch );
		return;
	}

	if ( opReplace ) {
		/* check for replacement text */
		if ( gtk_text_get_length ( GTK_TEXT (tReplace) ) ){
			replaceText  = gtk_editable_get_chars ( GTK_EDITABLE (tReplace),  0, -1);
		}
		else {
			msgBox ( _("Public enemy on the board!"),
					_("Shame on you!\nDon't you know it isn't nice to erase everything?"),
					_("Yeah, right!"));
			gtk_widget_grab_focus ( tReplace );
			return;
		}
	}
	
	/* it's ok, let's do it*/
	gtk_widget_set_sensitive ( bStop, TRUE);
	gtk_widget_set_sensitive ( bGo, FALSE);

	clearResultsList ();
	showResultsList ();
	addPatternInComboList ();

	location = gtk_entry_get_text ( GTK_ENTRY (comboBrowse_entry) );
	if ( !lstat ( location , &filestat ) ){
		if ( S_ISREG ( filestat.st_mode ) ){
			pattern = g_basename ( location );
			tmp = g_dirname ( location );
			location = g_strconcat ( tmp, "/", NULL );
			g_free ( tmp );
			have2freeLocation = TRUE;
			opRecursive = FALSE;
		}
	}
	if ( !have2freeLocation )
		pattern  = gtk_entry_get_text ( GTK_ENTRY (comboPattern_entry) );

	sbContext = gtk_statusbar_get_context_id ( GTK_STATUSBAR (sBar),
			"filelist" );
	pushStatus ( sbContext, _("Making file list") );
	gtk_progress_set_activity_mode ( GTK_PROGRESS (pBar), TRUE );
	gtk_progress_configure ( GTK_PROGRESS (pBar), 0, 0, 1 );
	forceGtkIdle ();
	
	/* create a working list of files */
	flist = getFileList ( location, pattern );
	
	pushStatus ( sbContext, _("Sorting file list...") );
	gtk_progress_set_activity_mode ( GTK_PROGRESS (pBar), FALSE );
	gtk_progress_bar_set_bar_style ( GTK_PROGRESS_BAR (pBar),
			GTK_PROGRESS_DISCRETE );
	gtk_progress_bar_update ( GTK_PROGRESS_BAR (pBar), 1 );
	forceGtkIdle ();
	flist = g_slist_sort ( flist, (GCompareFunc) sortFiles );
	wlist = flist;
	popStatus ( sbContext );
	
	/* remove "Making..." msg */
	popStatus ( sbContext );

	/* fill the results clist with the file list */
	gtk_progress_bar_update ( GTK_PROGRESS_BAR (pBar), 0 );
	
	/* now, that we have the file list, seek & destroy! */
	i = g_slist_length ( flist );
	if ( i ){
		gtk_progress_configure ( GTK_PROGRESS (pBar), 1, 0, i );
		gtk_progress_bar_set_bar_style ( GTK_PROGRESS_BAR (pBar),
			GTK_PROGRESS_CONTINUOUS );
		gtk_progress_set_format_string ( GTK_PROGRESS (pBar),
			_("File %v of %u") );
		gtk_progress_set_show_text ( GTK_PROGRESS (pBar), TRUE );
		i = 1;
		while ( wlist && notCanceled ){
			if ( doSearchReplace ( wlist, searchText, replaceText, i ) )
				i++;
			wlist = wlist -> next;
		}
	}

	/* let's free the list */
	wlist = flist;
	while ( wlist ){
		if ( wlist -> data ){
			g_free ( ((fileList*) (wlist -> data)) -> filename );
			g_free ( wlist -> data );
		}
		wlist = wlist -> next;
	}
	g_slist_free ( flist );
	
	gtk_progress_set_show_text ( GTK_PROGRESS (pBar), FALSE );
	gtk_progress_bar_update ( GTK_PROGRESS_BAR (pBar), 0 );
	
	gtk_widget_set_sensitive ( bStop, FALSE);
	gtk_widget_set_sensitive ( bGo, TRUE);

	g_free ( searchText );
	g_free ( replaceText );
	if ( have2freeLocation ){
		opRecursive = GTK_TOGGLE_BUTTON ( cRecursive ) -> active;
		g_free ( location );
	}
}

void cb_Stop( GtkWidget *w, gpointer data)
{
	notCanceled = FALSE;
	gtk_widget_set_sensitive ( bGo, TRUE);
	gtk_widget_set_sensitive ( bStop, FALSE);
}

void cb_Refresh( GtkWidget *w, gpointer data)
{
	g_print ("Refresh\n");
}

void cb_NewS( GtkWidget *w, gpointer data)
{
	GtkText *t;
	gushort i;

	for ( i = 0; i < 2; i++ ){
		if (i)
			t = GTK_TEXT ( tSearch );
		else
			t = GTK_TEXT ( tReplace );

		gtk_text_freeze ( t );
		gtk_text_set_point ( t, 0 );
		gtk_text_forward_delete ( t, gtk_text_get_length (t) );
		gtk_text_thaw ( t );
	}
	hideResultsList ();
}

void cb_switchpage( GtkWidget *n, gpointer data)
{
	int page = gtk_notebook_get_current_page( GTK_NOTEBOOK (n) );
	
	if ( page == clipboardPage ){
		GtkText *t = GTK_TEXT ( clipboard );
		GtkEditable *e = GTK_EDITABLE ( clipboard );
		
		gtk_text_freeze ( t );
		gtk_editable_set_editable ( e, TRUE );
		gtk_editable_delete_text ( e, 0, -1 );
		gtk_editable_set_position ( e, 0 );
		gtk_editable_paste_clipboard ( e );
		gtk_editable_set_editable ( e, FALSE );
		gtk_text_thaw ( t );
	}
}

void cb_Mail( GtkWidget *w, gpointer data)
{
	gchar *mailtext;
	
	mailtext = gtk_editable_get_chars( GTK_EDITABLE (data), 0, -1);
	if ( strlen (mailtext) )
		mail_it ( mailtext );
	else {
		msgBox ( _("Wow, the void!"),
				_("Gotcha!\n"
				"I bet you can't write. More:\n"
				"You can't even spell CAT.\n"
				"Don't you?"),
				_("CUT") );
		gtk_widget_grab_focus ( tMail );
	}
	g_free (mailtext);
}

void cb_Link( GtkWidget *w, gpointer data)
{
	open_uri ( "http://vbd.go.ro/software/turma/", NULL );
}

void cb_Arrow( GtkWidget *w, gpointer data)
{
	switchResultsList ();
}

